/*
 *   Copyright (C) 2009-2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Config.h"
#include "Globals.h"
#include "DMRIdleRX.h"

const q15_t SCALING_FACTOR1 = 10240;      // Q15(0.625)
const q15_t SCALING_FACTOR2 = 1365;       // Q15(1 / 24)

const uint8_t MAX_SYNC_SYMBOLS_ERRS = 2U;
const uint8_t MAX_SYNC_BYTES_ERRS   = 3U;

const uint8_t BIT_MASK_TABLE[] = {0x80U, 0x40U, 0x20U, 0x10U, 0x08U, 0x04U, 0x02U, 0x01U};

#define WRITE_BIT(p,i,b) p[(i)/8] = (b) ? (p[(i)/8] | BIT_MASK_TABLE[(i)%8]) : (p[(i)/8] & ~BIT_MASK_TABLE[(i)%8])

CDMRIdleRX::CDMRIdleRX() :
m_bitBuffer(),
m_buffer(),
m_bitPtr(0U),
m_dataPtr(0U),
m_endPtr(999U),
m_maxCorr(0),
m_centre(0),
m_threshold(0)
{
}

void CDMRIdleRX::samples(const q15_t* samples, uint8_t length)
{
  for (uint8_t i = 0U; i < length; i++)
    processSample(samples[i]);
}

void CDMRIdleRX::processSample(q15_t sample)
{
  m_bitBuffer[m_bitPtr] <<= 1;
  if (sample < 0)
    m_bitBuffer[m_bitPtr] |= 0x01U;

  m_buffer[m_dataPtr] = sample;

  if (countBits32((m_bitBuffer[m_bitPtr] & DMR_SYNC_SYMBOLS_MASK) ^ DMR_DATA_SYNC_SYMBOLS) <= MAX_SYNC_SYMBOLS_ERRS) {
    uint16_t ptr = (m_dataPtr + DMR_FRAME_LENGTH_SAMPLES - DMR_SYNC_LENGTH_SAMPLES + DMR_RADIO_SYMBOL_LENGTH - 1U) % DMR_FRAME_LENGTH_SAMPLES;

    q31_t corr = 0;
    q31_t mean = 0;
    q15_t max = -16000;

    uint32_t mask = 0x00800000U;
    for (uint8_t i = 0U; i < DMR_SYNC_LENGTH_SYMBOLS; i++, mask >>= 1) {
      bool b = (DMR_DATA_SYNC_SYMBOLS & mask) == mask;

      if (m_buffer[ptr] > max)
        max = m_buffer[ptr];

      corr += b ? -m_buffer[ptr] : m_buffer[ptr];
      mean += m_buffer[ptr];

      ptr = (ptr + DMR_RADIO_SYMBOL_LENGTH) % DMR_FRAME_LENGTH_SAMPLES;
    }

    if (corr > m_maxCorr) {
      q31_t v2 = mean * SCALING_FACTOR2;
      q15_t centre = q15_t(__SSAT(v2 >> 15, 16));

      q31_t v1 = (max - centre) * SCALING_FACTOR1;
      q15_t threshold = q15_t(v1 >> 15);

      uint8_t sync[DMR_SYNC_BYTES_LENGTH];
      uint16_t ptr = m_dataPtr - DMR_SYNC_LENGTH_SAMPLES + DMR_RADIO_SYMBOL_LENGTH;
      samplesToBits(ptr, DMR_SYNC_LENGTH_SYMBOLS, sync, 4U, centre, threshold);

      uint8_t errs = 0U;
      for (uint8_t i = 0U; i < DMR_SYNC_BYTES_LENGTH; i++)
        errs += countBits8((sync[i] & DMR_SYNC_BYTES_MASK[i]) ^ DMR_DATA_SYNC_BYTES[i]);

      if (errs <= MAX_SYNC_BYTES_ERRS) {
        DEBUG2("DMRIdleRX: Data sync found at", m_dataPtr);
        m_maxCorr   = corr;
        m_centre    = centre;
        m_threshold = threshold;
        m_endPtr    = (m_dataPtr + DMR_SLOT_TYPE_LENGTH_SAMPLES / 2U + DMR_INFO_LENGTH_SAMPLES / 2U - 1U) % DMR_FRAME_LENGTH_SAMPLES;
      }
    }
  }

  if (m_dataPtr == m_endPtr) {
    uint8_t frame[DMR_FRAME_LENGTH_BYTES + 1U];
    frame[0U] = 0x40U;      // Data sync

    uint16_t ptr = (m_endPtr + DMR_RADIO_SYMBOL_LENGTH) % DMR_FRAME_LENGTH_SAMPLES;
    samplesToBits(ptr, DMR_FRAME_LENGTH_SYMBOLS, frame, 8U, m_centre, m_threshold);

    serial.writeDMRData(frame, DMR_FRAME_LENGTH_BYTES + 1U);

    m_endPtr  = 999U;
    m_maxCorr = 0U;
  }

  m_dataPtr = (m_dataPtr + 1U) % DMR_FRAME_LENGTH_SAMPLES;
  m_bitPtr = (m_bitPtr + 1U) % DMR_RADIO_SYMBOL_LENGTH;
}

void CDMRIdleRX::samplesToBits(uint16_t start, uint8_t count, uint8_t* buffer, uint16_t offset, q15_t centre, q15_t threshold)
{
  for (uint8_t i = 0U; i < count; i++) {
    q15_t sample = m_buffer[start] - centre;

    if (sample < -threshold) {
      WRITE_BIT(buffer, offset, false);
      offset++;
      WRITE_BIT(buffer, offset, true);
      offset++;
    } else if (sample < 0) {
      WRITE_BIT(buffer, offset, false);
      offset++;
      WRITE_BIT(buffer, offset, false);
      offset++;
    } else if (sample < threshold) {
      WRITE_BIT(buffer, offset, true);
      offset++;
      WRITE_BIT(buffer, offset, false);
      offset++;
    } else {
      WRITE_BIT(buffer, offset, true);
      offset++;
      WRITE_BIT(buffer, offset, true);
      offset++;
    }

    start = (start + DMR_RADIO_SYMBOL_LENGTH) % DMR_FRAME_LENGTH_SAMPLES;
  }
}

uint8_t CDMRIdleRX::countBits8(uint8_t pattern) const
{
  uint8_t count;
  
  for (count = 0U; pattern; count++)
    pattern &= pattern - 1U;

  return count;
}

uint8_t CDMRIdleRX::countBits32(uint32_t pattern) const
{
  uint8_t count;
  
  for (count = 0U; pattern; count++)
    pattern &= pattern - 1U;

  return count;
}

