/*
 *   Copyright (C) 2013,2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Config.h"
#include "Globals.h"

#include "SerialPort.h"

const uint8_t MMDVM_FRAME_START  = 0xE0U;

const uint8_t MMDVM_GET_VERSION  = 0x00U;
const uint8_t MMDVM_GET_STATUS   = 0x01U;
const uint8_t MMDVM_SET_CONFIG   = 0x02U;
const uint8_t MMDVM_SET_MODE     = 0x03U;

const uint8_t MMDVM_CAL_DATA     = 0x08U;

const uint8_t MMDVM_DSTAR_HEADER = 0x10U;
const uint8_t MMDVM_DSTAR_DATA   = 0x11U;
const uint8_t MMDVM_DSTAR_LOST   = 0x12U;
const uint8_t MMDVM_DSTAR_EOT    = 0x13U;

const uint8_t MMDVM_DMR_DATA     = 0x18U;
const uint8_t MMDVM_DMR_LOST     = 0x19U;
const uint8_t MMDVM_DMR_SHORTLC  = 0x1AU;
const uint8_t MMDVM_DMR_START    = 0x1BU;

const uint8_t MMDVM_YSF_DATA     = 0x20U;
const uint8_t MMDVM_YSF_EOT      = 0x21U;
const uint8_t MMDVM_YSF_LOST     = 0x22U;

const uint8_t MMDVM_ACK          = 0x70U;
const uint8_t MMDVM_NAK          = 0x7FU;

const uint8_t MMDVM_DUMP         = 0xF0U;
const uint8_t MMDVM_DEBUG1       = 0xF1U;
const uint8_t MMDVM_DEBUG2       = 0xF2U;
const uint8_t MMDVM_DEBUG3       = 0xF3U;
const uint8_t MMDVM_DEBUG4       = 0xF4U;
const uint8_t MMDVM_DEBUG5       = 0xF5U;

const uint8_t HARDWARE[] = "MMDVM 20151015 (D-Star/DMR/System Fusion)";

const uint8_t PROTOCOL_VERSION   = 0x01U;

CSerialPort::CSerialPort() :
m_buffer(),
m_len(0U),
m_ptr(0U)
{
}

void CSerialPort::sendACK() const
{
  uint8_t reply[4U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 4U;
  reply[2U] = MMDVM_ACK;
  reply[3U] = m_buffer[2U];

  Serial.write(reply, 4);
}

void CSerialPort::sendNAK(uint8_t err) const
{
  uint8_t reply[5U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 5U;
  reply[2U] = MMDVM_NAK;
  reply[3U] = m_buffer[2U];
  reply[4U] = err;

  Serial.write(reply, 5);
}

void CSerialPort::getStatus() const
{
  uint8_t reply[11U];

  // Send all sorts of interesting internal values
  reply[0U]  = MMDVM_FRAME_START;
  reply[1U]  = 10U;
  reply[2U]  = MMDVM_GET_STATUS;

  reply[3U]  = 0x00U;
  if (dstarEnable)
    reply[3U] |= 0x01U;
  if (dmrEnable)
    reply[3U] |= 0x02U;
  if (ysfEnable)
    reply[3U] |= 0x04U;

  reply[4U]  = uint8_t(modemState);

  reply[5U]  = tx ? 0x01U : 0x00U;

  if (io.hasADCOverflow())
    reply[5U] |= 0x02U;

  if (io.hasBufferOverflow())
    reply[5U] |= 0x04U;

  if (dstarEnable)
    reply[6U] = dstarTX.getSpace();
  else
    reply[6U] = 0U;

  if (dmrEnable) {
    reply[7U] = dmrTX.getSpace1();
    reply[8U] = dmrTX.getSpace2();
  } else {
    reply[7U] = 0U;
    reply[8U] = 0U;
  }

  if (ysfEnable)
    reply[9U] = ysfTX.getSpace();
  else
    reply[9U] = 0U;

  Serial.write(reply, 10);
}

void CSerialPort::getVersion() const
{
  uint8_t reply[100U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_GET_VERSION;

  reply[3U] = PROTOCOL_VERSION;

  uint8_t count = 4U;
  for (uint8_t i = 0U; HARDWARE[i] != 0x00U; i++, count++)
    reply[count] = HARDWARE[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

uint8_t CSerialPort::setConfig(const uint8_t* data, uint8_t length)
{
  if (length < 7U)
    return 4U;

  bool rxInvert  = (data[0U] & 0x01U) == 0x01U;
  bool txInvert  = (data[0U] & 0x02U) == 0x02U;
  bool pttInvert = (data[0U] & 0x04U) == 0x04U;

  dstarEnable = (data[1U] & 0x01U) == 0x01U;
  dmrEnable   = (data[1U] & 0x02U) == 0x02U;
  ysfEnable   = (data[1U] & 0x04U) == 0x04U;

  txDelay = data[2U];
  if (txDelay > 100U)
    return 4U;

  modemState = MMDVM_STATE(data[3U]);

  if (modemState != STATE_IDLE && modemState != STATE_DSTAR && modemState != STATE_DMR && modemState != STATE_YSF && modemState != STATE_CALIBRATE)
    return 4U;
  if (modemState == STATE_DSTAR && !dstarEnable)
    return 4U;
  if (modemState == STATE_DMR && !dmrEnable)
    return 4U;
  if (modemState == STATE_YSF && !ysfEnable)
    return 4U;

  uint8_t rxLevel = data[4U];
  uint8_t txLevel = data[5U];

  uint8_t colorCode = data[6U];
  if (colorCode > 15U)
    return 4U;

  dmrTX.setColorCode(colorCode);

  io.setParameters(rxInvert, txInvert, pttInvert, rxLevel, txLevel);

  io.start();

  return 0U;
}

uint8_t CSerialPort::setMode(const uint8_t* data, uint8_t length)
{
  if (length < 1U)
    return 4U;

  modemState = MMDVM_STATE(data[0U]);

  if (modemState != STATE_IDLE && modemState != STATE_DSTAR && modemState != STATE_DMR && modemState != STATE_YSF && modemState != STATE_CALIBRATE)
    return 4U;
  if (modemState == STATE_DSTAR && !dstarEnable)
    return 4U;
  if (modemState == STATE_DMR && !dmrEnable)
    return 4U;
  if (modemState == STATE_YSF && !ysfEnable)
    return 4U;

  return 0U;
}

void CSerialPort::start()
{
  Serial.begin(115200);
}

void CSerialPort::process()
{
  while (Serial.available()) {
    uint8_t c = Serial.read();

    if (m_ptr == 0U && c == MMDVM_FRAME_START) {
      // Handle the frame start correctly
      m_buffer[0U] = c;
      m_ptr = 1U;
      m_len = 0U;
    } else if (m_ptr > 0U) {
      // Any other bytes are added to the buffer
      m_buffer[m_ptr] = c;
      m_ptr++;

      // Once we have enough bytes, calculate the expected length
      if (m_ptr == 2U)
        m_len = m_buffer[1U];

      if (m_ptr == 3U && m_len > 100U) {
        sendNAK(3U);
        m_ptr = 0U;
        m_len = 0U;
      }

      // The full packet has been received, process it
      if (m_ptr == m_len) {
        uint8_t err = 2U;

        switch (m_buffer[2U]) {
          case MMDVM_GET_STATUS:
            getStatus();
            break;

          case MMDVM_GET_VERSION:
            getVersion();
            break;

          case MMDVM_SET_CONFIG:
            err = setConfig(m_buffer + 3U, m_len - 3U);
            if (err == 0U)
              sendACK();
            else
              sendNAK(err);
            break;

          case MMDVM_SET_MODE:
            err = setMode(m_buffer + 3U, m_len - 3U);
            if (err == 0U)
              sendACK();
            else
              sendNAK(err);
            break;

          case MMDVM_CAL_DATA:
            if (modemState == STATE_CALIBRATE)
              err = calTX.write(m_buffer + 3U, m_len - 3U);
            if (err == 0U) {
            sendACK();
          } else {
            DEBUG2("Received invalid calibration data", err);
            sendNAK(err);
          }
          break;

          case MMDVM_DSTAR_HEADER:
            if (dstarEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_DSTAR)
                err = dstarTX.writeHeader(m_buffer + 3U, m_len - 3U);
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to D-Star");
                modemState = STATE_DSTAR;
              }
            } else {
              DEBUG2("Received invalid D-Star header", err);
              sendNAK(err);
            }
            break;

          case MMDVM_DSTAR_DATA:
            if (dstarEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_DSTAR)
                err = dstarTX.writeData(m_buffer + 3U, m_len - 3U);
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to D-Star");
                modemState = STATE_DSTAR;
              }
            } else {
              DEBUG2("Received invalid D-Star data", err);
              sendNAK(err);
            }
            break;

          case MMDVM_DSTAR_EOT:
            if (dstarEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_DSTAR)
                err = dstarTX.writeEOT();
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to D-Star");
                modemState = STATE_DSTAR;
              }
            } else {
              DEBUG2("Received invalid D-Star EOT", err);
              sendNAK(err);
            }
            break;

          case MMDVM_DMR_DATA:
            if (dmrEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_DMR)
                err = dmrTX.writeData(m_buffer + 3U, m_len - 3U);
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to DMR");
                modemState = STATE_DMR;
              }
            } else {
              DEBUG2("Received invalid DMR data", err);
              sendNAK(err);
            }
            break;

          case MMDVM_DMR_START:
            if (dmrEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_DMR)
                err = dmrTX.writeStart(m_buffer + 3U, m_len - 3U);
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to DMR");
                modemState = STATE_DMR;
              }
            } else {
              DEBUG2("Received invalid DMR start", err);
              sendNAK(err);
            }
            break;

          case MMDVM_DMR_SHORTLC:
            if (dmrEnable)
              err = dmrTX.writeShortLC(m_buffer + 3U, m_len - 3U);
            if (err != 0U) {
              DEBUG2("Received invalid DMR Short LC", err);
              sendNAK(err);
            }
            break;

          case MMDVM_YSF_DATA:
            if (ysfEnable) {
              if (modemState == STATE_IDLE || modemState == STATE_YSF)
                err = ysfTX.writeData(m_buffer + 3U, m_len - 3U);
            }
            if (err == 0U) {
              if (modemState == STATE_IDLE) {
                DEBUG1("Mode set to System Fusion");
                modemState = STATE_YSF;
              }
            } else {
              DEBUG2("Received invalid System Fusion data", err);
              sendNAK(err);
            }
            break;

          case MMDVM_YSF_EOT:
            if (ysfEnable)
              err = ysfRX.writeEOT();
            if (err != 0U) {
              DEBUG2("Received invalid System Fusion EOT Message", err);
              sendNAK(err);
            }
            break;

          default:
            // Handle this, send a NAK back
            sendNAK(1U);
            break;
        }

        m_ptr = 0U;
        m_len = 0U;
      }
    }
  }
}

void CSerialPort::writeDStarHeader(const uint8_t* header, uint8_t length)
{
  if (modemState != STATE_DSTAR && modemState != STATE_IDLE)
    return;

  if (!dstarEnable)
    return;

  uint8_t reply[50U];
  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DSTAR_HEADER;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = header[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

void CSerialPort::writeDStarData(const uint8_t* data, uint8_t length)
{
  if (modemState != STATE_DSTAR && modemState != STATE_IDLE)
    return;

  if (!dstarEnable)
    return;

  uint8_t reply[20U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DSTAR_DATA;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = data[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

void CSerialPort::writeDStarLost()
{
  if (modemState != STATE_DSTAR)
    return;

  if (!dstarEnable)
    return;

  uint8_t reply[3U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 3U;
  reply[2U] = MMDVM_DSTAR_LOST;

  Serial.write(reply, 3);
}

void CSerialPort::writeDStarEOT()
{
  if (modemState != STATE_DSTAR)
    return;

  if (!dstarEnable)
    return;

  uint8_t reply[3U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 3U;
  reply[2U] = MMDVM_DSTAR_EOT;

  Serial.write(reply, 3);
}

void CSerialPort::writeDMRData(const uint8_t* data, uint8_t length)
{
  if (modemState != STATE_DMR && modemState != STATE_IDLE)
    return;

  if (!dmrEnable)
    return;

  uint8_t reply[40U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DMR_DATA;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = data[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

void CSerialPort::writeDMRLost(uint8_t data)
{
  if (modemState != STATE_DMR)
    return;

  if (!dmrEnable)
    return;

  uint8_t reply[4U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 4U;
  reply[2U] = MMDVM_DMR_LOST;
  reply[3U] = data;

  Serial.write(reply, 4);
}

void CSerialPort::writeYSFData(const uint8_t* data, uint8_t length)
{
  if (modemState != STATE_YSF && modemState != STATE_IDLE)
    return;

  if (!ysfEnable)
    return;

  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_YSF_DATA;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = data[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

void CSerialPort::writeYSFLost()
{
  if (modemState != STATE_YSF)
    return;

  if (!ysfEnable)
    return;

  uint8_t reply[3U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 3U;
  reply[2U] = MMDVM_YSF_LOST;

  Serial.write(reply, 3);
}

void CSerialPort::writeCalData(const uint8_t* data, uint8_t length)
{
  if (modemState != STATE_CALIBRATE)
    return;

  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_CAL_DATA;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = data[i];

  reply[1U] = count;

  Serial.write(reply, count);
}

#if defined(WANT_DEBUG)

void CSerialPort::writeDump(const uint8_t* data, uint8_t length)
{
  uint8_t reply[200U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DUMP;

  uint8_t count = 3U;
  for (uint8_t i = 0U; i < length; i++, count++)
    reply[count] = data[i];

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeDebug(const char* text)
{
  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG1;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeDebug(const char* text, int16_t n1)
{
  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG2;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[count++] = (n1 >> 8) & 0xFF;
  reply[count++] = (n1 >> 0) & 0xFF;

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeDebug(const char* text, int16_t n1, int16_t n2)
{
  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG3;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[count++] = (n1 >> 8) & 0xFF;
  reply[count++] = (n1 >> 0) & 0xFF;

  reply[count++] = (n2 >> 8) & 0xFF;
  reply[count++] = (n2 >> 0) & 0xFF;

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeDebug(const char* text, int16_t n1, int16_t n2, int16_t n3)
{
  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG4;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[count++] = (n1 >> 8) & 0xFF;
  reply[count++] = (n1 >> 0) & 0xFF;

  reply[count++] = (n2 >> 8) & 0xFF;
  reply[count++] = (n2 >> 0) & 0xFF;

  reply[count++] = (n3 >> 8) & 0xFF;
  reply[count++] = (n3 >> 0) & 0xFF;

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeDebug(const char* text, int16_t n1, int16_t n2, int16_t n3, int16_t n4)
{
  uint8_t reply[130U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG5;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[count++] = (n1 >> 8) & 0xFF;
  reply[count++] = (n1 >> 0) & 0xFF;

  reply[count++] = (n2 >> 8) & 0xFF;
  reply[count++] = (n2 >> 0) & 0xFF;

  reply[count++] = (n3 >> 8) & 0xFF;
  reply[count++] = (n3 >> 0) & 0xFF;

  reply[count++] = (n4 >> 8) & 0xFF;
  reply[count++] = (n4 >> 0) & 0xFF;

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

void CSerialPort::writeAssert(bool cond, const char* text, const char* file, long line)
{
  if (cond)
    return;

  uint8_t reply[200U];

  reply[0U] = MMDVM_FRAME_START;
  reply[1U] = 0U;
  reply[2U] = MMDVM_DEBUG2;

  uint8_t count = 3U;
  for (uint8_t i = 0U; text[i] != '\0'; i++, count++)
    reply[count] = text[i];

  reply[count++] = ' ';
  
  for (uint8_t i = 0U; file[i] != '\0'; i++, count++)
    reply[count] = file[i];

  reply[count++] = (line >> 8) & 0xFF;
  reply[count++] = (line >> 0) & 0xFF;

  reply[1U] = count;

  Serial.write(reply, count);
  Serial.flush();
}

#endif

