/*
 *   Copyright (C) 2009-2015 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Config.h"
#include "Globals.h"
#include "YSFRX.h"

const unsigned int BUFFER_LENGTH = 200U;

const q15_t THRESHOLD = 2541;

const uint32_t PLLMAX = 0x10000U;
const uint32_t PLLINC = PLLMAX / YSF_RADIO_SYMBOL_LENGTH;
const uint32_t INC    = PLLINC / 32U;

const uint64_t DATA_SYNC_BITS = 0x000000D471C9634DU;
const uint64_t DATA_SYNC_MASK = 0x000000FFFFFFFFFFU;

const uint8_t DATA_SYNC_ERRS  = 3U;

const unsigned int MAX_SYNC_BITS = 4U * YSF_DATA_LENGTH_BITS;

const uint8_t BIT_MASK_TABLE[] = {0x80U, 0x40U, 0x20U, 0x10U, 0x08U, 0x04U, 0x02U, 0x01U};

#define WRITE_BIT(p,i,b) p[(i)/8] = (b) ? (p[(i)/8] | BIT_MASK_TABLE[(i)%8]) : (p[(i)/8] & ~BIT_MASK_TABLE[(i)%8])
#define READ_BIT(p,i)    (p[(i)/8] & BIT_MASK_TABLE[(i)%8])

CYSFRX::CYSFRX() :
m_pll(0U),
m_prev(false),
m_rxState(YSFRXS_NONE),
m_patternBuffer(0x00U),
m_rxBuffer(),
m_rxBufferBits(0U),
m_dataBits(0U)
{
}

void CYSFRX::samples(const q15_t* samples, uint8_t length)
{
  for (uint16_t i = 0U; i < length; i++) {
    bool bit = samples[i] < 0;

    if (bit != m_prev) {
      if (m_pll < (PLLMAX / 2U))
        m_pll += INC;
      else
        m_pll -= INC;
    }

    m_prev = bit;

    m_pll += PLLINC;

    if (m_pll >= PLLMAX) {
      m_pll -= PLLMAX;

      if (samples[i] < -THRESHOLD) {
        processBit(false);
        processBit(true);
      } else if (samples[i] < 0) {
        processBit(false);
        processBit(false);
      } else if (samples[i] < THRESHOLD) {
        processBit(true);
        processBit(false);
      } else {
        processBit(true);
        processBit(true);
      }
    }
  }
}

void CYSFRX::processBit(bool bit)
{
  m_patternBuffer <<= 1;
  if (bit)
    m_patternBuffer |= 0x01U;

  // Fuzzy matching of the data sync bit sequence
  if (countBits((m_patternBuffer & DATA_SYNC_MASK) ^ DATA_SYNC_BITS) <= DATA_SYNC_ERRS) {
#if defined(WANT_DEBUG)
    if (m_rxState == YSFRXS_DATA) {
      if (m_rxBufferBits != 40U)
        DEBUG2("YSFRX: Found data sync in data, unaligned", m_rxBufferBits);
      else
        DEBUG1("YSFRX: Found data sync in data");
    } else {
      DEBUG1("YSFRX: Found data sync in data");
    }
#endif
    m_dataBits = MAX_SYNC_BITS;
    m_rxBufferBits = 40U;

    if (m_rxState == YSFRXS_NONE) {
      m_rxState = YSFRXS_DATA;
      io.setDecode(true);
    }
  }

  // Nothing more to do if we're not receiving data
  if (m_rxState == YSFRXS_NONE)
    return;

  WRITE_BIT(m_rxBuffer, m_rxBufferBits, bit);
  m_rxBufferBits++;

  // We've not seen a data sync for too long, signal RXLOST and change to RX_NONE
  m_dataBits--;
  if (m_dataBits == 0U) {
    DEBUG1("YSFRX: Sync timed out, lost lock");
    io.setDecode(false);

    serial.writeYSFLost();

    m_rxState = YSFRXS_NONE;
    return;
  }

  // Send a data frame to the host if the required number of bits have been received, or if a data sync has been seen
  if (m_rxBufferBits == YSF_DATA_LENGTH_BITS) {
    m_rxBuffer[0U] = YSF_DATA_SYNC_BYTES[0U];
    m_rxBuffer[1U] = YSF_DATA_SYNC_BYTES[1U];
    m_rxBuffer[2U] = YSF_DATA_SYNC_BYTES[2U];
    m_rxBuffer[3U] = YSF_DATA_SYNC_BYTES[3U];
    m_rxBuffer[4U] = YSF_DATA_SYNC_BYTES[4U];

    serial.writeYSFData(m_rxBuffer, YSF_DATA_LENGTH_BYTES);

    // Start the next frame
    ::memset(m_rxBuffer, 0x00U, YSF_DATA_LENGTH_BYTES);
    m_rxBufferBits = 0U;
  }
}

uint8_t CYSFRX::writeEOT()
{
  if (m_rxState == YSFRXS_DATA)
    return 4U;

  m_rxState = YSFRXS_NONE;
  io.setDecode(false);

  return 0U;
}

uint8_t CYSFRX::countBits(uint64_t pattern) const
{
  uint8_t count;
  
  for (count = 0U; pattern; count++)
    pattern &= pattern - 1U;

  return count;
}

